#!/opt/VRTSperl/bin/perl
# $Copyrights: Copyright (c) 2014 Symantec Corporation.
# All rights reserved.
#
# THIS SOFTWARE CONTAINS CONFIDENTIAL INFORMATION AND TRADE SECRETS OF
# SYMANTEC CORPORATION.  USE, DISCLOSURE OR REPRODUCTION IS PROHIBITED
# WITHOUT THE PRIOR EXPRESS WRITTEN PERMISSION OF SYMANTEC CORPORATION.
#
# The Licensed Software and Documentation are deemed to be commercial
# computer software as defined in FAR 12.212 and subject to restricted
# rights as defined in FAR Section 52.227-19 "Commercial Computer
# Software - Restricted Rights" and DFARS 227.7202, "Rights in
# Commercial Computer Software or Commercial Computer Software
# Documentation", as applicable, and any successor regulations. Any use,
# modification, reproduction release, performance, display or disclosure
# of the Licensed Software and Documentation by the U.S. Government
# shall be solely in accordance with the terms of this Agreement.  $

eval 'exec /opt/VRTSperl/bin/perl -I /opt/VRTSvcs/lib -S $0 ${1+"$@"}'
        if 0;

use strict;
use warnings;
use ag_i18n_inc;
use Getopt::Long;
use FindBin '$Bin';
use constant MAX_SLEEP_TIME => 30;

our $size="20m";
our $tmpfs_dir="/VCStmpfs";
our $cur_dir=$Bin;
our $busybox="/sbin/busybox";
our $agent_type = "SANBootMonitor";
our $sh="sh";

my %opts;
my $rm ="/bin/rm";
my $mv="/bin/mv";
my $cp="/bin/cp";
my $ln="/bin/ln";
my $grep="/bin/grep";
my $awk="/bin/awk";
my $sed="/bin/sed";
my $sort = "/bin/sort";
my $mkdir="/bin/mkdir";
my $mount="/bin/mount";
my $vxfenadm="/sbin/vxfenadm";
my $gabconfig="/sbin/gabconfig";
my $vxfenrc="/etc/init.d/vxfen";
my $vcsrc="/etc/init.d/vcs";
my $vcs_lock="VRTSvcs/lock";
my $vcs_tmp="VRTSvcs/log/tmp";
my $vcs_ldf="VRTSvcs/ldf";
my $hasys="$VCSHOME/bin/hasys";
my $perl="/opt/VRTSperl/bin/perl";
my ($vcsnostart, $config, $start_vxfen) =("0", "0", "0");

my $vcshome = "$VCSHOME/bin";
my $vcslog = "$VCSLOG/log";
my $vcsconf= "$VCSCONF/conf";
my $maincf = "$vcsconf/config/main.cf";
my $sanboot="$vcshome/$agent_type";


printUsage() if($#ARGV > 1);

GetOptions('nostart'    => \$vcsnostart,
		'configure|C' => \$config,
		'unconfigure|U' => sub { unconfigureVCS() },
		'help|h'        => sub { printUsage() }
        ) or die("Error in parsing command line arguments\n");

# Main
if ($config == 1 || $vcsnostart == 1) {
	validate();
	setup();
	configureVCS() if($vcsnostart != 1); 
}
else {
	printUsage();
}

sub printUsage {

	print "\nUsage: $0 --nostart | --configure | --unconfigure [--help]\n\n";
        printf "%-20s : \%s", "--nostart ", " Does not start VCS after $agent_type configuration\n";
        printf "%-20s : \%s", "--configure|-C ", " Configure the $agent_type\n";
        printf "%-20s : \%s", "--unconfigure|-U ", " Unconfigure the $agent_type\n";
	printf "%-20s : \%s", "--help|-h ", " Print help message\n";

	exit(1);
}

sub validate {
	
	my ($stderr, $ret) =  (0) x 2;
	
	# Verify that platform is linux
	print "\n";
	printf "%-100s" , "[Verifying] Platform is Linux..";
	$ret = $^O;
	if ($ret !~ /Linux/i ) {
		printf "\%s", "Failed\n";
		print "\nERROR: Setup script can only run on Linux.\n";
		exit(1);
	}
	printf "\%s", "Done\n";
	# Verify that VCS is not running
	printf "%-100s" , "[Verifying] VCS is not running..";
	$stderr=qx($hasys -state);
	if( $stderr !~ /V-16-1-10600/ )	{
		printf "\%s", "Failed\n";
		print "\nERROR: VCS still running on the node. Please stop VCS manually.\n";
		exit(1);
	}
	printf "\%s", "Done\n";
	# Verify that VXFEN is not running
	printf "%-100s" , "[Verifying] Fencing is not running..";
	$ret=system("$vxfenadm -d > /dev/null 2>&1");
	if( $ret == 0 )	{
		printf "\%s", "Failed\n";
		print "\nERROR: Fencing still running on the node. Please stop Fencing manually.\n";
		exit(1);
	}
	printf "\%s", "Done\n";
	# Verify that busybox is installed on the system
	printf "%-100s", "[Verifying] Busybox is installed..";
	$ret=system("ls $busybox > /dev/null 2>&1");
	if( $ret != 0) {
		printf "\%s", "Failed\n";
		print "\nERROR: Busybox binary not found at '$busybox'.\n";
		exit(1);
	}
	printf "\%s", "Done\n";
	# Verify that busybox supports "sh" applet
	printf "%-100s", "[Verifying] Busybox supports shell..";
	$ret=`$busybox sh -? 2>&1`;
	if($ret =~ /applet not found/) {
		$ret=`$busybox ash -? 2>&1`;
		if($ret =~ /applet not found/) {
			printf "\%s", "Failed\n";
			print "\nERROR: Busybox does not support 'sh/ash' applet. Exiting.\n";
			exit(1);
		}
		$sh="ash";	
	}
	printf "\%s", "Done\n";
	#Verify that SANBootMonitor Agent is installed
	printf "%-100s", "[Verifying] $agent_type agent is present..";
	$ret=system("ls -d $sanboot > /dev/null 2>&1");
	if( $ret != 0) {
		printf "\%s", "Failed\n";
		print "\nERROR: $agent_type agent missing at $sanboot.\n";
		exit(1);
	}
	printf "\%s", "Done\n\n";
}

sub setup {
	
	my ($ret, $op, $tmp) = (undef) x 3;
	my @out = ();
	# Verify if tmpfs_dir is already on tmpfs. If so do not create and mount again
	if ( -e $tmpfs_dir) {
		printf "%-100s", "[Verifying] Is Ramfs already mounted at $tmpfs_dir..";
		($op) = qx($mount | $grep -w "$tmpfs_dir");
		if (defined $op) {
			@out = split(/\s+/, $op);
			if($out[-2] eq "tmpfs") {
				printf "\%s", "Done\n";
				print "NOTICE: Filesystem tmpfs already mounted at $tmpfs_dir. Skipping creation.\n";
			}
			else {
				printf "\%s", "Failed\n";
				print "\nERROR: Another filesystem already mounted at $tmpfs_dir. Unmount and retry.\n";
				exit(1);
			}
		}
		else {
			$ret=system("$mount -o size=$size -t tmpfs none $tmpfs_dir");
			if( $ret != 0) {
				printf "\%s", "Failed\n";
				print "\nERROR: Failed to create tmpfs at $tmpfs_dir.\n";
				exit(1);
			}
			printf "\%s", "Done\n";
		}
	}
	elsif(! -e $tmpfs_dir) {
		printf "%-100s", "[Creating] Mountpoint at $tmpfs_dir..";
		qx($mkdir -p $tmpfs_dir);
		printf "\%s", "Done\n";
		$ret=system("$mount -o size=$size -t tmpfs none $tmpfs_dir");
		printf "%-100s", "[Creating] Ramfs at $tmpfs_dir..";
		if( $ret != 0) {
			printf "\%s", "Failed\n";
			print "\nERROR: Failed to create tmpfs at $tmpfs_dir.\n";
			exit(1);
		}
		printf "\%s", "Done\n";
	}
	
	# Replace "sh" intepretor in the entry points
	$tmp = $tmpfs_dir;
	$tmp =~ s/\//\\\//g;
	my @sanboot_ep = qw (online offline clean monitor);
	for my $ep (@sanboot_ep) {
		qx($perl -pi -e 's/^#!.*/#!$tmp\\/busybox\\ $sh/g' $sanboot/$ep);
	}
	# Create VCS directories and symlinks
	create_vcsdirs_n_symlinks();

	printf "%-100s", "[Copying] $agent_type agent files..";
	qx($mkdir -p $tmpfs_dir/$agent_type);
	for my $ep (@sanboot_ep) {
		$ret = system("$cp -rf $sanboot/$ep $tmpfs_dir/$agent_type 2> /dev/null");
		if ($ret != "0") {
			printf "\%s", "Failed\n";
			print "\nERROR: Unable to copy $sanboot/$ep to $tmpfs_dir/$agent_type.\n";
			exit(1);
		}
	}
	printf "\%s", "Done\n";

	printf "%-100s", "[Copying] Busybox binary..";

	if(! -x "$tmpfs_dir/busybox") {
		$ret = system("$cp -rfL $busybox $tmpfs_dir 2> /dev/null");
		if ($ret != "0"){ 
			printf "\%s", "Failed\n";
			print "\nERROR: Unable to copy busybox binary to $tmpfs_dir\n"; 
			exit(1);
		}
		printf "\%s", "Done\n";
	}
	else { 
		printf "\%s", "Done\n";
		print "NOTICE: Busybox binary already present at $tmpfs_dir. Skipping copy.\n"; 
	}

# 	Copy init script and create links inside runlevels
	setup_runlevel(1);
# 	Create main.cf with new type SANBootMonitor
	modify_maincf();

	return;
}

sub create_vcsdirs_n_symlinks
{
	printf "%-100s", "[Creating] VCS directories inside $tmpfs_dir..";
	qx($mkdir -p $tmpfs_dir/$vcs_lock);
	qx($mkdir -p $tmpfs_dir/$vcs_tmp);
	qx($mkdir -p $tmpfs_dir/$vcs_ldf);
	printf "\%s", "Done\n";

	# Create symlinks for VCS directories
	printf "%-100s", "[Creating] Symlinks inside $VCSLOG...";

	my ($lock_dir, $logtmp_dir, $ldf_dir)  = ("$VCSLOG/lock", "$vcslog/tmp", "$VCSLOG/ldf");
	my ($lock_dir_bk, $logtmp_dir_bk, $ldf_dir_bk)  = ("$lock_dir.bk", "$logtmp_dir.bk", "$ldf_dir.bk");
	if (-d "$lock_dir") {
		qx($cp -rfL $lock_dir $lock_dir_bk 2> /dev/null);
                qx($rm -rf $lock_dir 2> /dev/null);
                qx($ln -s $tmpfs_dir/$vcs_lock $lock_dir 2> /dev/null);
                qx($cp -rfL $lock_dir_bk $tmpfs_dir/$vcs_lock 2> /dev/null);
                qx($rm -rf $lock_dir_bk 2> /dev/null);
        }
        if (-d "$logtmp_dir") {
		qx($cp -rfL $logtmp_dir $logtmp_dir_bk 2> /dev/null);
                qx($rm -rf $logtmp_dir 2> /dev/null);
                qx($ln -s $tmpfs_dir/$vcs_tmp $logtmp_dir 2> /dev/null);
                qx($cp -rfL $logtmp_dir_bk $tmpfs_dir/$vcs_tmp 2> /dev/null);
                qx($rm -rf $logtmp_dir_bk 2> /dev/null);
        }
	if (-d "$ldf_dir") {
		qx($cp -rfL $ldf_dir $ldf_dir_bk 2> /dev/null);
                qx($rm -rf $ldf_dir 2> /dev/null);
                qx($ln -s $tmpfs_dir/$vcs_ldf $ldf_dir 2> /dev/null);
	}
	printf "\%s", "Done\n";

	return;
}

sub setup_runlevel
{

	my @run_level = qw (3 4 5);
	my $sanboot_init_script = "/etc/init.d/vcssanboot";
	my $setup = shift;

	if($setup == 1){
	   printf "%-100s", "[Copying] Init scripts to make setup boot persistent..";
	   if ( ! -f $sanboot_init_script ) {
		printf "\%s", "Failed\n";
		print "\nERROR: Init script missing at $sanboot_init_script.";
		exit(1);
	  }
	}
	else {
	   printf "%-100s", "[Removing] vcssanboot init script..";
	}

	my $rl_init_dir =  "/etc"; # run level init directory
	qx($grep "SUSE Linux" /etc/issue);
	if ($? == 0) { #in SUSE Linux rc directories are under /etc/init.d
		$rl_init_dir .= "/init.d";
	}

	for my $rl (@run_level) {
		my $rl_snb_path = "$rl_init_dir/rc$rl.d/S65vcssanboot";
		if($setup == 1){
			qx($ln -s $sanboot_init_script $rl_snb_path 2>/dev/null) if(! -l $rl_snb_path);
		}
		else {
			qx(rm -f $rl_snb_path 2>/dev/null) if(-l $rl_snb_path);
		}
	}
	printf "\%s", "Done\n";

	return;
}

sub modify_maincf
{
	printf "%-100s", "[Modifying] main.cf with type $agent_type..";
	my $agentcf = $agent_type . "Type.cf";
	my $ret = system("$grep '$agentcf' $maincf > /dev/null 2>&1");
	if ($ret != 0) {
		printf "%-100s", "[Creating] main.cf on the node..";
		qx(echo 'include "$agentcf"' > /tmp/main.cf.tmp);
		qx($mv $maincf "$maincf.bk") if ( -e $maincf);
		qx($mv /tmp/main.cf.tmp $maincf);
		qx(cat "$maincf.bk" >> $maincf);
		printf "\%s", "Done\n";
	}
	printf "\%s", "Done\n";
	printf "%-100s", "[Verifying] main.cf on the node..";
        $ret = system("$vcshome/hacf -verify config 2> /dev/null");
        if($ret != 0 ) {
                printf "\%s", "Failed\n";
                print "\nERROR: Verification of $maincf file failed.";
                exit(1);
        }
	printf "\%s", "Done\n\n";
	return;
}

sub start_fencing
{
	my $host = shift;
	my ( $retry, $time ) = (6, 5);
	my $ret = qx($vxfenrc status 2> /dev/null);

	if($ret =~ /not loaded/) {	# start if not loaded
	   printf "%-100s", "[Starting] fencing on node $host..";
	   qx($vxfenrc start &>/dev/null);
	   printf "\%s", "Done\n";
	
	   printf "%-100s", "[Verifying] Fencing is running on the node..";
	   while($retry > 0) {
		$ret = qx($vxfenrc status 2> /dev/null);
		if($ret !~ /not loaded/) {
			printf "\%s", "Done\n";
			last;
		}
		else {
			$retry--;
			sleep $time;
		}
	   }

	   if($retry == 0){
		   printf "\%s", "Failed\n";
		   print "WARNING: Fencing failed to start. Check $vcslog/vxfen/vxfen.log for details.\n";
	   }
	} 
	return;
}

sub stop_fencing
{
	my $host = shift;
	my ( $retry, $time ) = (6, 5);
	my $ret = qx($vxfenrc status 2> /dev/null);

	if($ret !~ /not loaded/) {	# stop if loaded
		printf "%-100s", "[Stopping] fencing on node $host..";
		qx($vxfenrc stop &> /dev/null);
		printf "\%s", "Done\n";

		printf "%-100s", "[Verifying] Fencing is not running on the node..";
		while($retry > 0) {
			$ret = qx($vxfenrc status 2> /dev/null);
			if($ret =~ /not loaded/) {
				printf "\%s", "Done\n";
				$start_vxfen = 1;
				last;
			}
			else {
				$retry--;
				sleep $time;
			}
		}

		if($retry == 0){
			printf "\%s", "Failed\n";
			print "WARNING: Fencing failed to stop. Check $vcslog/vxfen/vxfen.log for details.\n";
		}
	}
	return;
}
  
sub start_gab
{
	printf "%-100s", "[Starting] GAB..";
	my $ret=qx($gabconfig -x -c &> /dev/null);
	if($? == 1) {
		printf "\%s", "Failed\n";
		print "ERROR: GAB failed to start $ret\n";
		exit(1);
	}
	printf "\%s", "Done\n";
	return;
}

sub stop_gab
{
	printf "%-100s", "[Stopping] GAB..";
	my $ret = qx($gabconfig -U &> /dev/null);
	if($? == 1) {
		printf "\%s", "Failed\n";
		print "ERROR: GAB failed to stop $ret\n";
		exit(1);
	}
	printf "\%s", "Done\n";
	return;
}

sub start_vcs
{
	my $host = shift;
	printf "%-100s", "[Starting] VCS on node $host..";
	qx($vcsrc start 2> /dev/null);
	sleep(10);
	my $ret=system("$vcshome/hasys -wait $host SysState RUNNING -time 60 2> /dev/null");
	if ($ret != 0) {
		printf "\%s", "Failed\n";
		print "\nERROR: VCS failed to start on the node. Start VCS manually or Check $vcslog/engine_A.log for details.\n";
		exit(1);
	}
	printf "\%s", "Done\n";
	return;
}

sub stop_vcs
{
	my $ret=qx($hasys -state);
	printf "%-100s", "[Stopping] VCS ...";
	if ($ret !~ /V-16-1-10600/) {
		qx($vcshome/hastop -all 2> /dev/null);
		my $try_count = MAX_SLEEP_TIME;
		while ($try_count > 0) {
			$ret=qx($hasys -state);
			last if ($ret =~ /V-16-1-10600/);
			$try_count--;
		}
		if ($try_count == 0) {
			printf "\%s", "Failed\n";
			print "\nERROR: VCS failed to stop on the node. Stop VCS manually and retry...\n";
			exit(1);
		}
	}
	qx($vcsrc stop &> /dev/null);
	printf "\%s", "Done\n";
	printf "%-100s", "[Stopping] CmdServer ...";
	qx($vcshome/CmdServer -stop);
	printf "\%s", "Done\n";
	return;
}

sub configureVCS {

	my $lltstat = "/sbin/lltstat";
	my $lltconfig = "/sbin/lltconfig";
	printf "%-100s", "[Verifying] LLT is running on the node..";
	my $ret = qx($lltconfig 2>&1);
	if ($ret =~ /not/ || $ret =~ /ERROR/) {
		printf "\%s", "Failed\n";
		print "\nERROR: LLT is not running on the node.\n";
		exit(1);
	}
	printf "\%s", "Done\n";
	my $hostname = qx($lltstat -H);
	chomp($hostname);

	$ret = qx($grep -c "UseFence.*SCSI3" $maincf 2> /dev/null);
	if ($ret =~ /1/) {
		start_fencing($hostname);
	}
	start_vcs($hostname);

	# Check if SANBootMonitor Resource type has been added
	$ret = system("$vcshome/hatype -display $agent_type > /dev/null 2>&1");
	if ($ret != 0 ) {
		print "\nERROR:  Agent type $agent_type not found in VCS configuration.\n";
		printf "%-100s", "Hint: Is it possible that VCS was running on some other node of the cluster with older configuration? Ensure that VCS is stopped on all nodes before reruning the script.";
		exit(1);
	}

	configure_grp($hostname);
	return;
}

sub configure_grp
{
	my $hostname = shift;
	my $grp_name = "sanboot_grp";
	my $res_name = "sanboot_res";
	my ($ret, $sysList, $autoList) = ("") x 3 ;
	my @nodes = qx($sort /etc/llthosts | $awk '{print \$NF}');
	chomp(@nodes);

	for(my $i = 0; $i <= $#nodes; $i++) {
		$sysList .= "$nodes[$i] $i ";
		$autoList .= "$nodes[$i] ";
	}
	qx($vcshome/haconf -makerw 2> /dev/null);
	qx($vcshome/hatype -modify $agent_type AgentDirectory "$tmpfs_dir/$agent_type");
	printf "%-100s", "[Verifying] Is Group $grp_name configured..";
	$ret = qx($vcshome/hagrp -state $grp_name 2>&1);
	printf "\%s", "Done\n";
	if($ret =~ /V-16-1-40131/) {
		printf "%-100s", "[Adding] Group $grp_name to VCS configuration..";
		qx($vcshome/hagrp -add $grp_name);
		qx($vcshome/hagrp -modify $grp_name SystemList $sysList);
		qx($vcshome/hagrp -modify $grp_name AutoStartList $autoList);
		qx($vcshome/hagrp -modify $grp_name Parallel 1);
		printf "\%s", "Done\n";
	}
	else {
		print "NOTICE: $grp_name already in configuration.\n";
	}
	printf "%-100s", "[Verifying] Is Resource $res_name configured..";
	$ret = qx($vcshome/hares -state $res_name 2>&1);
	printf "\%s", "Done\n";
	if($ret =~ /V-16-1-40130/) {
		printf "%-100s", "[Adding] Resource $res_name to VCS configuration..";
		qx($vcshome/hares -add $res_name $agent_type $grp_name);
		qx($vcshome/hares -modify $res_name MountPoint \"$tmpfs_dir\");
		qx($vcshome/hares -modify $res_name Enabled 1);
		printf "\%s", "Done\n";
	}
	else {
		print "NOTICE: $res_name already in configuration.\n";
	}
	sleep(3);
	qx($vcshome/haconf -dump -makero);
	qx($vcshome/hares -probe $res_name -sys $hostname > /dev/null 2>&1);
	printf "%-100s", "[Attempting] to online $grp_name on $hostname..";
	sleep(5);
	qx($vcshome/hagrp -online $grp_name -sys $hostname > /dev/null 2>&1);
	$ret=system("$vcshome/hagrp -wait $grp_name State ONLINE -sys $hostname -time 60 > /dev/null 2>&1");
	if ($ret != 0) {
		printf "\%s", "Failed\n";
		print "\nERROR: Group $grp_name failed to online on the node. Check $vcslog/engine_A.log for details.\n";
		exit(1);
	}
	printf "\%s", "Done\n";
}

sub delete_grp{

	my ($grp_name, $res_name) = ("sanboot_grp" , "sanboot_res");
	printf "%-100s", "[Removing] Group $grp_name from VCS configuration..";
	qx($vcshome/haconf -makerw 2> /dev/null);
	qx($vcshome/hares -delete $res_name 2> /dev/null);
	qx($vcshome/hagrp -delete  $grp_name 2> /dev/null);
	qx($vcshome/haconf -dump -makero 2> /dev/null);
	qx($sed -i '/$agent_type/d' $maincf 2> /dev/null);
	printf "\%s", "Done\n";
	return;
}

sub unconfigureVCS
{
	print "Unconfiguration will stop VCS, fencing and GAB on node. Do you want to proceed? (yes/no):";
	my $unconfig = "";
	while (1) {
		$unconfig = <>;
		chomp($unconfig);
		last if($unconfig eq "yes");
		exit(1) if ($unconfig eq "no");
		print "(yes/no):";
	}

	print "Unconfiguring $agent_type agent ...\n";
	my $lltstat = "/sbin/lltstat";
	my @out = ();
	my $hostname = qx($lltstat -H);
	chomp($hostname);
#	Remove SANBootMonitor group from main.cf configuration.
	delete_grp();
#	Stop VCS if running
	stop_vcs();
#       Stop fencing  if running
	stop_fencing($hostname);
#	Stop gab
	stop_gab();
	sleep(5); # waiting for gablogd to stop

#	Remove the symbolic links created by the setup script
	my ($lock_dir, $logtmp_dir, $ldf_dir)  = ("$VCSLOG/lock", "$vcslog/tmp", "$VCSLOG/ldf");
	qx($rm $lock_dir  $logtmp_dir  $ldf_dir 2> /dev/null);
#	Unmount the tmpfs mount point if present
	if( -e $tmpfs_dir ) {
		my ($op) = qx($mount | $grep -w "$tmpfs_dir");
		if (defined $op) {
			printf "%-100s", "[Unmounting] $tmpfs_dir..";
			@out = split(/\s+/, $op);
			if($out[-2] eq "tmpfs") { #unmount if type is tmpfs
				qx(umount $tmpfs_dir 2> /dev/null);
				if($? != 0) {
					printf "\%s", "Failed\n";
					print "ERROR: Failed to unmount $tmpfs_dir. Unmount $tmpfs_dir manually and retry...\n";
					exit(1);
				}
				qx($rm -rf $tmpfs_dir &>/dev/null);
				printf "\%s", "Done\n";
			}
			else{
				printf "\%s", "Done\n";
				print "NOTICE: Another filesystem mounted at $tmpfs_dir. Skipping unmount\n";
			}
		}
	}
# 	Delete links inside runlevels
	setup_runlevel(0);
	printf "\%s", "Unconfiguration of $agent_type agent completed successfully\n";

#	Now restart the VCS services.
	start_gab();
	start_fencing($hostname) if ($start_vxfen);
	start_vcs($hostname);
	print "\nNOTE: Run $0 --unconfigure on all cluster node to unconfigure $agent_type agent\n";
exit(1);
}

